#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <fcntl.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include "getopt.h"
#include "sockx.h"
#include "signalx.h"
#include "parser.h"
#include "debug.h"

//Timeout times and buffer sizes
#define INTRAMESSAGETO  60
#define POKEMASTERTO    (5*60)
#define DLOADBUFSIZE    2048

//Network-related parameters
int   MASTERPORT        = 2345;
char *MASTERNAME        = "localhost";
int   LOADERPORT        = 0;
char *LOCALAPPNAME      = "./topcjob";    //Download binary on donor machine
char *CHECKPOINTAPPNAME = "./checkpoint"; //Checkpoint and download binary
char *ORIGINALAPPNAME   = "./a.out";      //Original binary

//Max and min
#define max(a, b) ((a) > (b) ? (a) : (b))
#define min(a, b) ((a) < (b) ? (a) : (b))

//Variables
int ssock;


/*Get acknowledge. Returns NULL if ACK, or error message if NACK. The
 *error message is formed according to the REASON string or, if such
 *is missing, accordingly
 */
//Silly pretty (as opposed to the usual adjective ugly used in this
//context) kludges
char *syntaxError  = "syntax error";
char *unknownError = "unspecified error";

//This gets reply to command
char *GetAck(int fd, int timeout) {
  const char syntax[] = "| ACK | NACK REASON";
  treerec *tree;
  char *res;
  int ID;

  for (;;) {
    tree = fdparse(fd, timeout, syntax);
    if (tree == NULL)  //Unparseable
      return syntaxError;
    switch (tree->ID) {
      case 1:  //Empty line
        continue;
      case 2:  //ACK
        free_parse_tree(tree);
        return NULL;
      case 3:  //NACK
        res = tree->argv[0].text;
        if (res) res = strdup(res);
        free_parse_tree(tree);
        if (res) return res;
        return unknownError;
    }
  }
}


/*Check if any new slaves have connected. Returns the number of
 *such. Though this version uses select(), therefore can only
 *return either 0 or 1
 */
int poll_new_slaves() {
  int res;
  res = fdwaitr(ssock, 0);
  assert(res >= 0);  //There shouldn't be any errors, really:)
  return res;
}


/*In the Emacs distribution, where unexec() comes from originally,
 *it's only used with default section addresses, hence, if non-
 *default values are used, it refuses to work on some platforms
 */
int do_unexec() {
  unexec(CHECKPOINTAPPNAME, ORIGINALAPPNAME, 0, 0, 0);
}


/*This is where most of the work is done. Accept connections from
 *attaching slaves, download the checkpoint binary, and start it.
 *Returns the number of successfully attached slaves, which could
 *be either 0 or 1
 */
int attach_new_slaves() {
  char buffer[DLOADBUFSIZE];
  const char syntax[] = "ATTACH PORT";
  treerec *tree;
  char *name;
  int isock, osock, port, fd, size, len;

  //Accept connection from loader
  isock = Accept(ssock);
  if (isock < 0) return 0;

  //Receive ATTACH message
  tree = fdparse(isock, INTRAMESSAGETO, syntax);
  if (tree == 0) goto error1;
  if (tree->argv[0].text == NULL) goto error2;
  port = atoi(tree->argv[0].text);
  if (port < 0 || port > USHRT_MAX) goto error2;
  name = RemoteHost(isock);

  //Acknowledge ATTACH message
  DBGP_TREE(syntax, tree);
  free_parse_tree(tree);
  fdprintf(isock, INTRAMESSAGETO, "ACK\n");
  close(isock);

  //Connect to loader, download and start the binary
  osock = Socket(name, port);
  if (osock < 0) {
    DBGP("Can't connect to %s:%n\n", name, port);
    return 0;
  }

  //Open the binary
  fd = open(CHECKPOINTAPPNAME, O_RDONLY);
  if (fd < 0) {
    close(osock);
    DBGP("Can't open %s\n", CHECKPOINTAPPNAME);
    return 0;
  }
  size = lseek(fd, 0, SEEK_END);
  assert(size >= 0);
  lseek(fd, 0, SEEK_SET);

  //Download the binary
  fdprintf(osock, INTRAMESSAGETO, "DLOAD SIZE=%d\n", size);
  while ((len = read(fd, buffer, sizeof(buffer))) > 0)
    if (write(osock, buffer, len) != len) goto error4;
  close(fd);
  if (GetAck(osock, INTRAMESSAGETO) != NULL) goto error3;

  //Start the binary
  fdprintf(osock, INTRAMESSAGETO, "START\n", size);
  if (GetAck(osock, INTRAMESSAGETO) != NULL) {
    close(osock);
    DBGP("Can't start new slave on %s:%d\n", name, port);
    return 0;
  }
  close(osock);
  DBGP("Started new slave on %s:%d\n", name, port);
  return 1;

error4:
  close(fd);
error3:
  close(osock);
  DBGP("Can't download to %s:%n\n", name, port);
  return 0;

error2:
  free_parse_tree(tree);
error1:
  close(isock);
  DBGP("Received bad attach request\n");
  return 0;
}


//Test program entry point
int slave_counter = 0;

int main(int argc, char **argv) {
  if (strcmp(argv[0], LOCALAPPNAME) == 0) {
    printf("Job %d running\n", slave_counter);
    for (;; sleep(1)) {printf("."); fflush(stdout);}
  }

  ORIGINALAPPNAME = argv[0];
  ssock = ServerSocket(MASTERPORT);
  assert(ssock >= 0);
  DBGP("Master listening on %d\n", LocalPort(ssock));
  for (;; sleep(2)) {
    if (poll_new_slaves()) {
      ++slave_counter;
      do_unexec();
      attach_new_slaves();
    }
  }
  return 0;
}