#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <stdarg.h>
#include "parser.h"

//Line buffer sizes
#define INPUTLINEMAX  1024
#define OUTPUTLINEMAX 1024

//Max and min
#define max(a, b) ((a) > (b) ? (a) : (b))
#define min(a, b) ((a) < (b) ? (a) : (b))

//Compare two strings delimited by characters from a set
static int strtokcmp(const char *s1, const char *s2,
                     const char *delim1, const char *delim2) {
  return strncasecmp(s1, s2, max(strcspn(s1, delim1), strcspn(s2, delim2)));
}

//String operations
#define strskip(s, accept) ((s) + strspn((s), (accept)))
#define strfind(s, reject) ((s) + strcspn((s), (reject)))


//Wait for input data; timeout is in seconds, where -1 waits indefinitely
int fdwaitr(int fd, int timeout) {
  struct timeval tv;
  fd_set set;
  int res;

  do {
    FD_ZERO(&set);
    FD_SET(fd, &set);
    if (timeout != -1) {
      tv.tv_sec = timeout;  //Some system's select() adjusts these, some not...
      tv.tv_usec = 0;       //...so we choose the dirty and reliable way
      res = select(fd + 1, &set, NULL, NULL, &tv);
    }
    else res = select(fd + 1, &set, NULL, NULL, NULL);
  } while (res == -1 && errno == EINTR);
  return res;
}


//The parser routine
treerec *fdparse(int fd, int timeout, const char *syntax) {
  char line[INPUTLINEMAX + 1];
  treerec *tree;
  const char *word, *next, *cmd, *arg;
  int i, k, n, len;

  //Read in line
  for (n = 0;;) {
    if (n == INPUTLINEMAX) return NULL;
    if (fdwaitr(fd, timeout) != 1) return NULL;
    if (read(fd, line + n, 1) != 1) return NULL;
    if (line[n] == '\n') break;
    if (line[n] != '\r') ++n;
  }
  line[n] = '\0';


  //Find command in syntax definition
  word = strskip(line, " ");
  cmd = strskip(syntax, " ");
  
  for (i = 1;; ++i) {
    if (cmd == NULL) return NULL;
    if (strtokcmp(word, cmd, " ", " |") == 0) break;
    
    //Skip to the beginning of the next command, delimited by '|'
    cmd = strchr(cmd, '|');
    if (cmd == NULL) return NULL;
    cmd = strskip(cmd + 1, " ");
  }
  
  //Scan syntax definition, counting up command arguments
  arg = strskip(strfind(cmd, " |"), " ");
  for (n = 0;; ++n) {
    switch (*arg) {
      case '\0':
      case '|': goto done1;
      case '(': arg = strskip(strchr(arg, ')') + 1, " "); break;
      default: arg = strskip(strfind(arg, " |"), " "); break;
    }
  }
done1:

  //Allocate memory for parse "tree"
  tree = calloc(1, sizeof(treerec));
  assert(tree != NULL);
  tree->ID = i;
  if (n > 0) {
    tree->argn = n;
    tree->argv = calloc(n, sizeof(argrec));
    assert(tree->argv != NULL);
  }


  //Now parse the arguments
  word = strskip(strfind(word, " "), " ");
  cmd = strskip(strfind(cmd, " |"), " ");
  
  for (;*word != '\0' ;) {  
    arg = cmd;
    for (k = 0;; ++k) {
      switch (*arg) {
        //Unrecognized argument, ignore it
        case '\0':
        case '|':
          word = strskip(strfind(word, " ="), " ");  //Skip keyword
          if (*word == '=') {
            word = strskip(word + 1, " ");
            if (*word == '"') {  //Skip quoted...
	      word = strchr(word + 1, '"');
              if (word == NULL) goto error1;
            }
            else word = strfind(word, " ");  //...or unquoted text
          }
          word = strskip(word, " ");  //Find next argument
          goto done2;
    
        //See if it's a selector argument  
        case '(':
          arg = strskip(arg + 1, " ");
          for (i = 1; *arg != ')'; ++i) {
            if (strtokcmp(word, arg, " ", " )") == 0) {
              tree->argv[k].sel = i;                    //Record the selector
              word = strskip(strfind(word, " "), " ");  //Get next argument
              goto done2;
            }
            arg = strskip(strfind(arg, " )"), " ");  //Scan syntax definition
          }
          arg = strskip(arg + 1, " ");
          break;

        //Must be a text argument
        default:
          if (strtokcmp(word, arg, " =", " |") == 0) {
            word = strskip(strfind(word, " ="), " ");  //Find the equal sign
            if (*word != '=') goto error1;  //Make sure equal sign present
            word = strskip(word + 1, " ");  //Find text string
	  
            if (*word == '"') {  //String may be either quoted...
              ++word;
	      next = strchr(word, '"');
              if (next == NULL) goto error1;
              len = next - word;
              ++next;
            }
            else {  //...or unquoted
              next = strfind(word, " ");
              len = next - word;
            }

            if (tree->argv[k].text == NULL) {  //Prevent overwrite
              tree->argv[k].text = malloc(len + 1);  //Copy text string
              assert(tree->argv[k].text != NULL);
              strncpy(tree->argv[k].text, word, len);
              tree->argv[k].text[len] = '\0';
            }

            word = strskip(next, " ");  //Point at next argument
            goto done2;
          }
          arg = strskip(strfind(arg, " |"), " ");
          break;
      }
    }
  done2:
  }
  return tree;

error1:
  free_parse_tree(tree);
  return NULL;
}


//Free parse "tree"
void free_parse_tree(treerec *tree) {
  int i;
  if (tree != NULL) {
    if (tree->argv != NULL)
      for (i = 0; i < tree->argn; ++i)
        if (tree->argv[i].text != NULL)
          free(tree->argv[i].text);
    free(tree);
  }
}


//Print parse "tree"
#define fputn(str, n, fp) {\
  const char *ptr = (str);\
  int i;\
  for (i = (n); i > 0 && *ptr != 0; --i) fputc(*ptr++, fp);}

void print_parse_tree(const char *syntax, treerec *tree, FILE *fp) {
  const char *arg;
  int i, j;

  //Check for empty tree
  if (tree == NULL) {
    fputs("(null)", fp);
    return;
  }

  //Print command ID
  arg = strskip(syntax, " ");
  for (i = 1; i < tree->ID; ++i) {
    arg = strchr(arg, '|');
    if (arg == NULL) {
      fputs("Invalid command", fp);
      return;
    }
    arg = strskip(arg + 1, " ");
  }
  fputn(arg, strcspn(arg, " |"), fp);
  fputc(' ', fp);

  //Print arguments
  arg = strskip(strfind(arg, " |"), " ");
  for (i = 0;; ++i) {
    switch (*arg) {
      case '\0':
      case '|': return;

      case '(':
        arg = strskip(arg + 1, " ");
        if (tree->argv[i].sel > 0) {
          for (j = 1; j < tree->argv[i].sel; ++j)
            arg = strskip(strfind(arg, " )"), " ");
          fputn(arg, strcspn(arg, " )"), fp);
          fputc(' ', fp);
        }
        arg = strskip(strchr(arg, ')') + 1, " ");
        break;

      default:
        if (tree->argv[i].text != NULL) {
          fputn(arg, strcspn(arg, " |"), fp);
          fputc('=', fp);
          if (strchr(tree->argv[i].text, ' ') != NULL) {
            fputc('"', fp);
            fputs(tree->argv[i].text, fp);
            fputs("\" ", fp);
          }
          else {
            fputs(tree->argv[i].text, fp);
            fputc(' ', fp);
          }
        }
        arg = strskip(strfind(arg, " |"), " ");
        break;
    }
  }
}


//Print formatted string to a file descriptor
int fdprintf(int fd, int timeout, const char *format, ...) {
  char buffer[OUTPUTLINEMAX + 1];
  va_list args;
  struct timeval tv;
  fd_set set;
  int res, len;

  if (timeout >= 0) {
    tv.tv_sec = timeout;
    tv.tv_usec = 0;
    FD_ZERO(&set);
    FD_SET(fd, &set);
    do {
      res = select(fd + 1, NULL, &set, NULL, &tv);
    } while (res == -1 && errno == EINTR);
    if (res < 1) return res;
  }

  //Print the output
  va_start(args, format);
  vsprintf(buffer, format, args);
  len = strlen(buffer);
  res = write(fd, buffer, len);
  va_end(args);
  return len == res ? res : EOF;
}


//Debug code
#ifdef PARSER_MAIN

void putch(int ch) {
  putchar(ch);
  fflush(stdout);
}

void main() {
  char *syntax = "ack (start stop) | nack (attach)|dload name size";
  treerec *tree;
  tree = fdparse(0, 20, syntax);
  print_parse_tree(syntax, tree, stdout);
  printf("\n");
}

#endif
