/********************************************************************
 ********************************************************************
 ** USAGE:  topcc [ --seq | --mpi | --pthread ] THIS_FILE
 ** ALGORITHM:  Euclidean sieve for factorization
 ** shared data:  num_to_factor (divided by each factor found)
 ** task input:   input (test factor for num_to_factor)
 ** task output:  1 or 0 (true or false)
 ** update:  Divide num_to_factor by num
 ********************************************************************
 ********************************************************************/

#include "topc.h"
#include <stdlib.h> // for atol() [ Note long is only 32 bits on many mach's]

#ifndef TASK_INTERVAL
#define TASK_INTERVAL 1000
#endif
/* do AGGLOM_COUNT tasks at once, save on network bandwidth */
#ifndef AGGLOM_COUNT
#define AGGLOM_COUNT 1
#endif

// Used if TOPC_OPT_trace == 2, or: ./a.out --TOPC-trace=2
void trace_input( long *input ) { printf("%ld", *input); } 
void trace_result( long *input, long *output ) {
  if (*output == 1) printf("TRUE");
  if (*output == 0) printf("FALSE");
}
// This is equivalent to:  long task_out;
// But for SMP (shared memory), it also ensures that task_out is not shared.
// TOP-C implements only a single thread-private variable:  TOPC_private_global
typedef long TOPC_private_global_t;
#define task_out TOPC_private_global
TOPC_private_global_t task_out_dbg() {  /* needed for debugging in gdb */
  return task_out;
}

long factors[1000], max_factor=1, num_to_factor, last_num;
void MasSlaveFactor(long);

int main(int argc, char *argv[])
{
  int i;
  char *arg = argv[1];

  // Set up defaults.  Can be overridden, e.g.:  ./a.out --TOPC-trace=0
  TOPC_OPT_trace_input = trace_input;
  TOPC_OPT_trace_result = trace_result;
  TOPC_OPT_trace = 2;
  TOPC_init(&argc, &argv);
  // Important:  Inspect command line only after TOPC_init()
  if (argc == 1 ) {
    if (TOPC_is_master())
      printf("WARNING: Missing number arg on command line.  Using 123456789\n");
    arg = "123456789";
  }

  MasSlaveFactor( atol(arg) );
  if ( TOPC_is_master() ) {
    for (i = 1; i <= factors[ 0 ]; i++) printf("%ld ", factors[i]);
    printf("\n"); }
  TOPC_finalize();
  // printf("%d: Exiting\n", TOPC_rank()); fflush(stdout);
  exit(0);
}

TOPC_BUF GenerateTaskInput() {
  last_num = last_num + TASK_INTERVAL;
  if ( last_num > num_to_factor ) { return NOTASK; }
  return TOPC_MSG(&last_num, sizeof(last_num));
}

TOPC_BUF DoTask( long *input )
{
  long limit;
  long num = *input;

  limit = num + TASK_INTERVAL;
  for ( ; num < limit; num++ )
    if (num_to_factor % num == 0) {
      task_out = 1;
      return TOPC_MSG(&task_out, sizeof(long));
    }
  task_out = 0;
  return TOPC_MSG(&task_out, sizeof(long));
}

TOPC_ACTION CheckTaskResult( long *input, long *output )
{
  if ( *output == 0 ) return NO_ACTION;
  if ( *input < max_factor ) return UPDATE;
  if ( ! TOPC_is_up_to_date() ) return REDO;
  return UPDATE;
}

void UpdateSharedData( long *input, long *output )
{
  long limit, num = *input;
  int i;
  limit = num + TASK_INTERVAL;
  // NOTE: Suppose we factor 12, 4 is reported as factor before 2 is reported
  for ( ; num < limit; num++ ) {
    if ( num_to_factor % num == 0 ) {
      if ( num < max_factor ) {
        max_factor = 1; // re-compute max_factor
        for( i = 1; i <= factors[0]; i++ ) {
          if (factors[i] > max_factor) max_factor = factors[i];
          while (factors[i] % num == 0) {
            factors[i] = factors[i] / num;
            num_to_factor = num_to_factor * num;
          }
        }
      }
      while (num_to_factor % num == 0) {
        factors[ ++factors[0] ] = num;
        num_to_factor = num_to_factor / num;
        if (num_to_factor == 1) return;
      }
    }
  }
}

void MasSlaveFactor( long num )
{
  factors[0]=0;
  last_num = 2 - TASK_INTERVAL;
  num_to_factor = num;
  TOPC_master_slave(GenerateTaskInput, DoTask, CheckTaskResult,
                    UpdateSharedData);
}
