 /**********************************************************************
  * TOP-C (Task Oriented Parallel C)                                   *
  * Copyright (c) 2000 Gene Cooperman <gene@ccs.neu.edu>               *
  *                    and Victor Grinberg <victor@ccs.neu.edu>        *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact Gene Cooperman          *
  * <gene@ccs.neu.edu>.                                                *
  **********************************************************************/

/**********************************************************************
 * If you port this to another MPI implementation, be aware that this
 * code assumes that the rank of the initial (master) process is 0, and the
 * slave processes have contiguous ranks beginning at 1.
 * You will also need to change the include file (mpinu.h) below to your
 * mpi.h include file, and change the Makefile.in to use your libmpi.a
 * or libmpi.so in src/Makefile.in and in bin/topc-config.in
 * Then re-execute ./configure .
 * Check `./configure --help' to see if I yet offer
 *   ./configure --with-mpi-prefix=<DIR>
 **********************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <stdlib.h> // malloc()
#include <unistd.h> // alarm()
#include <signal.h> // signal(), pthread_kill()
#include <assert.h>
#include <mpinu.h>  // MPINU implementation of MPI
#include "comm.h"

#ifdef HAVE_PTHREAD
# include <pthread.h>
# ifdef HAVE_SEMAPHORE_H
#  include <semaphore.h> // Solaris:  needs -lposix4
# else
#  include "sem-pthread.h"
# endif
static pthread_t cache_thread;
#endif

//Compile-time parameters (see comm.h for more parameters)
#define NUM_MSG_CACHE_PTR 1000  // Number of pending msg's that can be cached

/***********************************************************************
 ***********************************************************************
 ** Communications functions, implemented over MPI
 ***********************************************************************
 ***********************************************************************/

/***********************************************************************
 * Initialize underlying communications library
 */
static void alarm_handler(int sig) {
  printf("TOP-C:  60 minutes elapsed without reply by master\n"
         "  slave exiting\n");
  exit(1);
}
static int use_alarm = 1;
void COMM_init(int *argc, char ***argv, struct opt_list *options_list) {
  int flag;
  MPI_Initialized(&flag);
  if (!flag) {
    MPI_Init(argc, argv);
    if (TOPC_OPT_slave_timeout == 0 || COMM_rank() == 0) // no alarm on master
      use_alarm = 0;
    if (use_alarm) signal( SIGALRM, alarm_handler );
    if (use_alarm) alarm(TOPC_OPT_slave_timeout);
    COMM_rank();  // pre-compute rank as static local var.
#ifdef HAVE_PTHREAD
    if (TOPC_OPT_safety < 14)
      cache_thread = pthread_self();
#endif
  }
}

/***********************************************************************
 * Shutdown underlying communications library
 */
void COMM_finalize() {
  int flag;
  MPI_Initialized(&flag);
  if (flag) {
    int rank = COMM_rank();
#ifdef HAVE_PTHREAD
    if (TOPC_OPT_safety < 14)
      // cache_thread is still self if COMM_receive() never called
      if (rank != 0 && ! pthread_equal(pthread_self(), cache_thread) ) {
        // MPI_Finalize() will close sockets while cache_thread could
        //   try to read from them, kill cache_thread now
        pthread_kill( cache_thread, SIGKILL );
        pthread_join( cache_thread, NULL );
      }
#endif
    MPI_Finalize();
    if (rank != 0) exit(0); /* If slave, exit */
  }
}

/***********************************************************************
 * Test if underlying communications library is running
 */
TOPC_BOOL COMM_initialized() {
  int flag;
  MPI_Initialized(&flag);
  return flag;
}

/***********************************************************************
 * Return boolean, true if master and slave share memory
 */
int COMM_is_shared_memory() {
  return 0;  // false
}

/***********************************************************************
 * Get number of nodes
 */
int COMM_node_count() {
  int count;
  MPI_Comm_size(MPI_COMM_WORLD, &count);
  return count;
}

/***********************************************************************
 * Get rank of this node
 */
int COMM_rank() {
  static int rank = -1; // -1 means don't know; static so we remember rank
  if (rank == -1 && COMM_initialized())
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  // Else in mpinu, absence of "-p4amslave" in argv[] indicates it's master
  // In our case, we can't be sure
  if ( rank == -1 ) {
    fprintf(stderr, "TOPC_rank()/TOPC_is_master() called"
       " before TOPC_init() in dist. memory model\n"
       "  Can't determine if this is master or slave in this version of MPI.\n"
       "Try placing TOPC_init() earlier"
       " ( before call to TOPC_rank()/TOPC_is_master() )\n");
    exit(1);
  }
  return rank;
}

/***********************************************************************
 * Set via SOFT_ABORT_TAG
 */
volatile int COMM_abort_pending = 0; // read on slave by TOPC_abort_pending()
// called on master by TOPC_abort()
void COMM_soft_abort() {
  int slave, num_slaves = COMM_node_count()-1;
  for (slave = 1; slave <= num_slaves; ++slave)
    COMM_send_msg(NULL, 0, slave, SOFT_ABORT_TAG);
  }

/***********************************************************************
 * Check if a message from a slave is instantly available (non-blocking)
 */
TOPC_BOOL COMM_probe_msg() {
  MPI_Status status;
  int flag;

  MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status);
  return flag;
}

/***********************************************************************
 * Check for a message from a slave (blocking)
 */
TOPC_BOOL COMM_probe_msg_blocking(int rank) {
  MPI_Status status;

  // Allow one hour (default) for master to reply and for our task, or we die.
  if (use_alarm) alarm(TOPC_OPT_slave_timeout);
  MPI_Probe(rank, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
  return 1;
}

/***********************************************************************
 * Send message
 */
void COMM_send_msg(void *msg, size_t msg_size, int dst, enum TAG tag) {
  MPI_Send(msg, (int)msg_size, MPI_BYTE, dst, tag, MPI_COMM_WORLD);
}

/***********************************************************************
 * malloc/free_receive_buffer tries to use our own buffer for
 * messages until msg_cache is full.  Then uses malloc.
 * Currently no buffer wraparound until cache_head = cache_tail.
 */
static char msg_cache[MSG_CACHE_SIZE];
static char *cache_head = msg_cache, *cache_tail = msg_cache;
static int msg_cache_free = MSG_CACHE_SIZE; // recorded, but not curr. used
#define ROUNDUP(x) ( (((x)+sizeof(int)-1) / sizeof(int)) * sizeof(int) )

//msg_cache has repeated:  {tag,size,msg} where size is size of msg.
static void *malloc_receive_buffer(int size) {
  void *buf;
  if (cache_tail + ROUNDUP(size)+sizeof(enum TAG)+sizeof(int)
      <= msg_cache + MSG_CACHE_SIZE) {
    msg_cache_free -= ( ROUNDUP(size)+sizeof(enum TAG)+sizeof(int) );
    *(enum TAG *)cache_tail = PING_TAG; // storing tag here is obsolete
    ((enum TAG *)cache_tail)++;
    *(int *)cache_tail = size;
    ((int *)cache_tail)++;
    buf = cache_tail;
    cache_tail += ROUNDUP(size);
    assert( cache_tail <= msg_cache + MSG_CACHE_SIZE );
  }
  else if (size > 0) {
    buf = malloc(size);
  }
  else
    buf = NULL;
  if (buf == NULL && size > 0)
    { fprintf(stderr, "TOP-C:  No memory for receive buffer.\n"); exit(1); }
  return buf;
}

static void free_receive_buf(void *msg) {
  if ((char *)msg >= msg_cache && (char *)msg <= msg_cache + MSG_CACHE_SIZE) {
    /* msg == cache_tail possible w/ message size of 0 */
    assert( (char *)msg >= cache_head+sizeof(enum TAG)+sizeof(int)
            && (char *)msg <= cache_tail );
    (char *)msg -= ( sizeof(enum TAG) + sizeof(int) );
    assert( *(enum TAG *)msg != NO_TAG );
    *(enum TAG *)msg = NO_TAG; // free msg
    if (msg == cache_head) {
      while (*(enum TAG *)cache_head == NO_TAG && cache_head != cache_tail) {
        cache_head += sizeof(enum TAG);
        msg_cache_free +=
          ( sizeof(enum TAG)+sizeof(int)+ROUNDUP(*(int *)cache_head) );
        cache_head += ( sizeof(int) + ROUNDUP(*(int *)cache_head) );
      }
      assert( cache_head <= cache_tail );
      if (cache_head == cache_tail)
        cache_head = cache_tail = msg_cache;
    }
  }
  else
    free(msg);
}

/***********************************************************************
 * Receive message into malloc'ed memory (blocking call)
 * receive from anybody, and set each of four param's if non-NULL
 */
//The fields of struct msg_ptr correspond to the args of COMM_receive_msg()
struct msg_ptr {
  void *msg;
  size_t msg_size;
  int src;
  enum TAG tag;
  int inuse;
} msg_cache_ptr[NUM_MSG_CACHE_PTR];
static volatile struct msg_ptr *cache_ptr_head = msg_cache_ptr,
                               *cache_ptr_tail = msg_cache_ptr;
#ifdef HAVE_PTHREAD
static sem_t cache_semaphore;
static sem_val = 0; // used by DEBUG
#endif

static void cache_msg() {
  struct msg_ptr msg_info;
  MPI_Status status;
  int size;

  do {
    MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
    MPI_Get_count(&status, MPI_BYTE, &size);
    //if size_t 64 bits, cast 32 bit int back to size_t
    msg_info.msg_size = size;

    if (status.MPI_TAG == SOFT_ABORT_TAG) {
      COMM_abort_pending = 1;
      assert(size == 0);
      MPI_Recv( NULL, 0, MPI_BYTE,
                status.MPI_SOURCE, SOFT_ABORT_TAG, MPI_COMM_WORLD, &status );
    }
  } while (status.MPI_TAG == SOFT_ABORT_TAG);

  // Add hook here if want options for caller to pass in buffer
  msg_info.msg = malloc_receive_buffer(size);
  MPI_Recv( msg_info.msg, size, MPI_BYTE,
            status.MPI_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
  msg_info.src            = status.MPI_SOURCE;
  msg_info.tag            = (enum TAG)status.MPI_TAG;
  assert( (msg_info.inuse = 1) > 0 );

#ifndef HAVE_PTHREAD
  assert(cache_ptr_tail == cache_ptr_head);
#endif
  *(cache_ptr_tail++) = msg_info;
  if (cache_ptr_tail == msg_cache_ptr + NUM_MSG_CACHE_PTR)
    cache_ptr_tail = msg_cache_ptr;
#ifdef HAVE_PTHREAD
  if (TOPC_OPT_safety < 14)
    if (COMM_rank() != 0) {
#     ifdef DEBUG
      printf("rank: %d, thread: %d posting (tag: %d); in queue: %d\n",
        COMM_rank(), pthread_self(), msg_info.tag, ++sem_val);
#     endif
      if (cache_ptr_tail == cache_ptr_head) {
        printf("TOP-C:  Overflow:  More than %d pending messages cached.\n"
               "        Re-compile src/comm-mpi.c with larger value"
                        " for NUM_MSG_CACHE_PTR\n",
               NUM_MSG_CACHE_PTR);
        abort();
      }
      sem_post(&cache_semaphore);
    }
#endif
}

#ifdef HAVE_PTHREAD
static void *cache_messages(void *dummy) {
  while (1)
    cache_msg();
  /* NOTREACHED */
  return NULL;
}
static void cache_messages_init() {
  static pthread_attr_t slave_attr;
  int val;

  val = sem_init( &cache_semaphore, 0, 0 );
  if (val == -1) { perror("TOP-C:  sem_init"); exit(1); }
  pthread_attr_init(&slave_attr);
  pthread_attr_setdetachstate(&slave_attr, PTHREAD_CREATE_JOINABLE);
#ifdef DEBUG
  printf("rank: %d, thread created\n", COMM_rank());
#endif
  val = pthread_create( &cache_thread, &slave_attr, cache_messages, NULL );
  if (val != 0) { fprintf(stderr, "%s\n", strerror(val) ); exit(1); }
}
#endif

void COMM_receive_msg(void **msg, size_t *msg_size, int *src, enum TAG *tag) {
  struct msg_ptr msg_info;
  static int cache_messages_inited = 0;

  // Allow one hour (default) for master to reply and for our task, or we die.
  if (use_alarm) alarm(TOPC_OPT_slave_timeout);
#ifdef HAVE_PTHREAD
  if (TOPC_OPT_safety < 14 && COMM_rank() != 0) {
    /* The actual receive is done asynchronously by a separate thread */
    if (! cache_messages_inited) {
      cache_messages_init();
      cache_messages_inited = 1;
    }
#   ifdef DEBUG
    printf("rank: %d, thread: %d waiting\n", COMM_rank(), pthread_self());
#   endif
    sem_wait(&cache_semaphore);
#   ifdef DEBUG
    printf("rank: %d, thread: %d done waiting\n", COMM_rank(), pthread_self());
#   endif
  }
  else
    cache_msg(); // Get and cache a message ourselves, if no thread
#else
  /* The actual receive is done by cache_msg() */
  cache_msg(); // Get and cache a message ourselves, if no thread
#endif

  msg_info = *(cache_ptr_head++);
  if (cache_ptr_head == msg_cache_ptr + NUM_MSG_CACHE_PTR)
    cache_ptr_head = msg_cache_ptr;

  if (msg) *msg            = msg_info.msg;
  if (msg_size) *msg_size  = msg_info.msg_size;
  if (src) *src            = msg_info.src;
  if (tag) *tag            = msg_info.tag;
#ifdef HAVE_PTHREAD
# ifdef DEBUG
  printf("rank: %d, thread: %d   received tag: %d\n; in queue: %d",
    COMM_rank(), pthread_self(), msg_info.tag, --sem_val);
# endif
#endif
  assert( --(msg_info.inuse) == 0 );
}

// balances malloc(), etc. in COMM_receive_msg()
void COMM_free_receive_buf(void *msg) {
  free_receive_buf(msg);
}

/***********************************************************************
 ***********************************************************************
 * SMP-related functions (e.g.:  private (non-shared) variables)
 ***********************************************************************
 ***********************************************************************/

void *COMM_private_global(size_t size) { // trivial for non-shared memory
  static void *private_global = NULL;
  if ( private_global == NULL ) {
    private_global = malloc(size);
    if (size >= sizeof(NULL))
      *(void **)private_global = NULL;
  }
  return private_global;
}
void COMM_begin_atomic_read() {};
void COMM_end_atomic_read() {};
void COMM_begin_atomic_write() {};
void COMM_end_atomic_write() {};
TOPC_BOOL COMM_is_in_atomic_section() { return 0; };
