#define DEBUG
 /**********************************************************************
  * TOP-C (Task Oriented Parallel C)                                   *
  * Copyright (c) 2000 Gene Cooperman <gene@ccs.neu.edu>               *
  *                    and Victor Grinberg <victor@ccs.neu.edu>        *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact Gene Cooperman          *
  * <gene@ccs.neu.edu>.                                                *
  **********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "comm.h"

#ifndef MSG_CACHE_SIZE
// This value should be same as default of ./configure --help
#define MSG_CACHE_SIZE 50000
#endif

static char msg_cache[MSG_CACHE_SIZE];
static char *cache_head = msg_cache, *cache_tail = msg_cache;
static int msg_cache_free = MSG_CACHE_SIZE; // recorded, but not curr. used

/***********************************************************************
 * Access functions for a message
 * A message is stored contiguously in msg_cache as:
 *                                  ( tag, source, size, buffer )
 */
#define ROUNDUP(x) ( (((x)+sizeof(int)-1) / sizeof(int)) * sizeof(int) )
#define sizeof_enum_TAG (ROUNDUP(sizeof(enum TAG)))
#define sizeof_msg_header ( sizeof_enum_TAG + 2*sizeof(int) )
#define sizeof_msg(size) ( sizeof_msg_header + ROUNDUP(size) )

#define MEM_buf_tag(buf) (*(enum TAG *)(buf - sizeof_enum_TAG - 2*sizeof(int)))
#define MEM_buf_source(buf) (*(int *)(buf - 2*sizeof(int)))
#define MEM_buf_size(buf) (*(int *)(buf - sizeof(int)))

/***********************************************************************
 * malloc/free_receive_buffer tries to use our own buffer for
 * messages until msg_cache is full.  Then uses malloc.
 * Note msg vs. buf in code:   msg = ( tag, source, size, buf )
 * Currently no buffer wraparound until cache_head = cache_tail.
 */
// MEM_malloc(int size, int source, enum TAG tag) {
static void *malloc_receive_buffer(int size, int source, enum TAG tag) {
  void *buf;
  if (cache_tail + sizeof_msg(size) <= msg_cache + MSG_CACHE_SIZE) {
    msg_cache_free -= sizeof_msg(size);
    buf = cache_tail;
    cache_tail += sizeof_msg(size);
    assert( cache_tail <= msg_cache + MSG_CACHE_SIZE );
  }
  else {
    buf = malloc(sizeof_msg(size));
    if (buf == NULL && size > 0) {
      fprintf(stderr, "TOP-C:  No memory to malloc a new buffer.\n");
      exit(1);
    }
  }

  buf += sizeof_msg_header;
  MEM_buf_tag(buf) = tag;
  MEM_buf_source(buf) = source;
  MEM_buf_size(buf) = size;

  return buf;
}

// MEM_free(char *buf) {
static void free_receive_buf(char *buf) {
  /* buf==cache_tail, or msg_cache + MSG_CACHE_SIZE possible if msg size is 0*/
  assert( buf >= cache_head + sizeof_msg_header && buf <= cache_tail );
  assert( MEM_buf_tag(buf) > NO_TAG && MEM_buf_tag(buf) <= MAX_TAG );
  assert( MEM_buf_source(buf) >= 0 && MEM_buf_source(buf) <= 100000 );
  MEM_buf_tag(buf) = NO_TAG; // free buf
  buf -= sizeof_msg_header;

  if (buf >= msg_cache && buf <= msg_cache + MSG_CACHE_SIZE) {
    if (buf == cache_head) {
      while (*(enum TAG *)cache_head == NO_TAG && cache_head != cache_tail) {
        cache_head += sizeof(enum TAG);
        msg_cache_free +=
          ( sizeof(enum TAG)+sizeof(int)+ROUNDUP(*(int *)cache_head) );
        cache_head += ( sizeof(int) + ROUNDUP(*(int *)cache_head) );
      }
      assert( cache_head <= cache_tail );
      if (cache_head == cache_tail)
        cache_head = cache_tail = msg_cache;
    }
  }
  else
    free(buf);
}
