/*  ISSUE:  In options.c, clean up standalone test */

/***********************************************************************
 ***********************************************************************
 ** USAGE:  get_topc_options( int *argc, char ***argv );
 **   Command line options:  --TOPC_XXX= (variable:  TOPC_OPT_XXX)
 ***********************************************************************
 ***********************************************************************/

/*  gcc -DTEST options.c; ./a.out --TOPC_help   for testing */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/stat.h>
#include "topc.h"
#include "comm.h"

#ifdef TEST
int COMM_is_shared_memory(void) { return 1; }
int COMM_node_count()
  {extern int TOPC_OPT_num_slaves; return TOPC_OPT_num_slaves + 1;};
#define UNINITIALIZED -1;
#endif

/* Defaults; types should also be declared extern in topc.h and comm.h */
#define bool int
int TOPC_OPT_num_slaves = UNINITIALIZED;
int TOPC_OPT_slave_wait = 0; /* seconds */
int TOPC_OPT_slave_timeout = 3600;
int TOPC_OPT_trace = 2;
// agglomeration in experimental version
// int TOPC_OPT_agglom_count = 1;
bool TOPC_OPT_help = 0;     /* --TOPC_help (no arg) sets it to 1 */
bool TOPC_OPT_verbose = 1;  /* --TOPC_verbose (no arg) sets it to 2 */
char *TOPC_OPT_procgroup = "./procgroup";
char *TOPC_OPT_topc_log = "-"; /* - = stdout, but should set to topc.log if is "" */
int TOPC_OPT_safety = 0;

enum opt_type { BOOL, INT, STRING };

struct option {
  char *option;
  void *var_ptr;
  enum opt_type type;
  char *help_string;
};

/* COULD COMPUTE DEFAULT NUM_SLAVES FIRST */

static struct option cmd_opt[] = {
  { "--TOPC_help", &TOPC_OPT_help, BOOL, "display this message" },
  { "--TOPC_verbose", &TOPC_OPT_verbose, INT, "[0..2]: set verbose mode" },
  { "--TOPC_num_slaves", &TOPC_OPT_num_slaves, INT,
    "number of slaves (sys-defined default)" },
  { "--TOPC_slave_wait", &TOPC_OPT_slave_wait, INT,
    "secs before slave starts (use w/ gdb attach)"},
  { "--TOPC_slave_timeout", &TOPC_OPT_slave_timeout, INT,
    "dist mem: secs to die if no msgs, 0=never"},
  { "--TOPC_trace", &TOPC_OPT_trace, INT,
    "trace (0: notrace, 1: trace, 2: user trace fncs.)" },
// agglomeration in experimental version
//  { "--TOPC_agglomeration", &TOPC_OPT_agglomeration, STRING,
//    "NOT IMPL:  agglomeration count" },
  { "--TOPC_procgroup", &TOPC_OPT_procgroup, STRING, "procgroup file (--mpi)" },
  { "--TOPC_topc_log", &TOPC_OPT_topc_log, STRING,
                       "NOT IMPL:  log file for TOPC output (\"-\" = stdout)" },
  { "--TOPC_safety", &TOPC_OPT_safety, INT,
    "[0..20]: higher turns off optimizations,\n"
    "                           see --TOPC_verbose" },
  { "", NULL, BOOL, "" }
};

#define COL2 27
#define COL3 55

static char *print_arg_type(enum opt_type type) {
  switch (type) {
  case BOOL: return "[=<0/1>]";
  case INT: return "=<int>";
  case STRING: return "=<string>";
  default: return "";
  }
}

static void print_help(char *command) {
  struct option *opt;
  int len1, len2;
  printf("\nUsage:  %s [TOPC_OPTION ...] [OTHER_OPTIONS ...]\n", command);
  printf("  where TOPC_OPTION is one of:\n");
  for ( opt = cmd_opt; opt->option[0] != '\0'; opt++ ) {
    printf("%s%s%n", opt->option, print_arg_type(opt->type), &len1 );
    printf("%*s%s %n", COL2 - len1, " ", opt->help_string, &len2 );
    if ( len1 + len2 > COL3 ) printf("\n%*s", COL3, " ");
    else printf("%*s", COL3 - len1 - len2, " ");
    switch (opt->type) {
    case BOOL:
      printf("[default: false]\n");
      break;
    case INT:
      printf("[default: %d]\n", *(int *)(opt->var_ptr));
      break;
    case STRING:
      printf("[default: \"%s\"]\n", *(char **)(opt->var_ptr));
      break;
    default:
      printf("*** FATAL:  invalid option type:  %d\n", opt->type); exit(1);
    }
  }
  printf("\nYou can change the defaults in the application source code.\n");
  printf("  For example, to change the default to --TOPC_trace=0,\n");
  printf("  add `TOPC_OPT_trace = 0;' before `TOPC_init(&argc, &argv);'\n");
}

static void set_option( struct option *opt, char *val ) {
  char *endptr; // Used to check validity of strtol() arg.
  switch (opt->type) {
  case BOOL:
    if ( val[0] == '\0' )
      val = NULL;
    else if ( val[0] == '=' )
      val = val + 1;
    else
      printf("*** WARNING:  %s: Boolean option takes no arguments.\n",
	     opt->option);
    break;
  case INT:
    if (val[0] == '\0' && strcmp("--TOPC_verbose", opt->option) == 0) {
      *(int *)(opt->var_ptr) = 2;
      return;
    }
  case STRING:
    if ( val[0] != '=' ) {
      printf( "*** FATAL:  %s=<val>:  option takes %s argument.\n",
	      opt->option, (opt->type==INT ? "an int" : "a string") );
      exit(1);
    }
    else
      val = val + 1; // val was "=<arg>", now val is "<arg>"
    break;
  }
  switch (opt->type) {
  case BOOL:
    if ( ! val )
      *(int *)(opt->var_ptr) = 1;
    break;
  case INT:
    *(int *)(opt->var_ptr) =
      strtol( val, &endptr, 10 );
    if (*endptr != '\0') {
      printf("Invalid integer value for %s=%s\n", opt->option, val);
      exit(1);
    }
    break;
  case STRING:
    *(char **)(opt->var_ptr) = malloc( strlen( val ) );
    strcpy( *(char **)(opt->var_ptr), val );
    break;
  }
}

static void get_options(int *argc, char ***argv) {
  struct option *opt;
  int i, j, end_options = 0; 
#ifdef TEST
printf("BEGIN:  *argc=%d\n", *argc);
#endif
  for ( i = 1, j = 1; j < *argc; j++ ) {
#ifdef TEST
printf("(*argv)[%d]=%s\n", j, (*argv)[j]);
#endif
    if ( strncmp( "--", (*argv)[j], 3) == 0 )
      end_options = 1;
    if ( ! end_options && strncmp( (*argv)[j], "--TOPC", 6 ) == 0 ) {
      for ( opt = cmd_opt; opt->option[0] != '\0'; opt++ ) {
        if ( strncmp( opt->option, (*argv)[j], strlen(opt->option) ) == 0 ) {
          set_option( opt, (*argv)[j] + strlen(opt->option) );
          if ( opt->var_ptr == &TOPC_OPT_procgroup ) {
	    char *old_str = (*argv)[j];
            (*argv)[j] = malloc( strlen((*argv)[j])
			+ strlen("--TOPC_procgroup") - strlen("-p4pg") + 1 );
	    strcpy( (*argv)[j], "-p4pg" );
	    strcpy( (*argv)[j] + strlen("-p4pg"),
		     old_str + strlen("--TOPC_procgroup") );
	    // Replace --TOPC_OPT_procgroup= by -p4pg= in (*argv)[j]
	    // BUG:  WHY?
	    //argv = memcpy( malloc((*argc+1)*sizeof(**argv)), argv,
            //               (*argc+1)*sizeof(**argv) );
            //(*argv)[j] += strlen("--TOPC_procgroup")-strlen("-p4pg");
            //memcpy( (*argv)[j], "-p4pg", strlen("-p4pg") );
            //(*argv)[i] = (*argv)[j];
            //i++;
          }
          break;
        }
      }
      if ( opt->option[0] == '\0' ) { // if "--TOPC" did not match anything
        printf("\n*** Invalid option:  %s\n\n", (*argv)[j]);
        print_help(**argv);
        exit(1);
      }
    }
    else {
      (*argv)[i] = (*argv)[j];
      i++;
    }
  }
  (*argv)[i] = (*argv)[j]; // Copy final null pointer
  *argc = i;
#ifdef TEST
printf("END:  *argc=%d\n", *argc);
#endif
}

#define ARGS_MAX 100
static char *saved_args[ARGS_MAX]; /* from argv[] */
static time_t start_time;
// These are defined in procgroup.c
void TOPC_OPT_pre_extend_procgroup
  (char *procgroup_file, int num_slaves, int *argc, char ***argv);
void TOPC_OPT_post_extend_procgroup(int *argc, char ***argv);

void TOPC_OPT_pre_init( int *argc, char ***argv ) {
  int i;
  struct stat buf;

  for (i = 0; i <= *argc && i < ARGS_MAX; i++)
    saved_args[i] = (*argv)[i];

  get_options(argc, argv);

  if (TOPC_OPT_safety == 19)
    TOPC_OPT_num_slaves = 2;
  if (TOPC_OPT_safety >= 20)
    TOPC_OPT_num_slaves = 1;

  if (TOPC_OPT_help) {
    print_help( **argv ); // **argv == command name ($0 on command line)
    exit( 0 );
  }

#ifndef TEST
  /*If dist. mem., and can't find procgroup, skip it.  This could be a slave.*/
  if ( ! COMM_is_shared_memory()
       && 0 == stat( TOPC_OPT_procgroup, &buf ) )
    TOPC_OPT_pre_extend_procgroup(TOPC_OPT_procgroup, TOPC_OPT_num_slaves,
				  argc, argv);
#endif
}

void TOPC_OPT_post_init( int *argc, char ***argv ) {
#ifndef TEST
  TOPC_OPT_post_extend_procgroup(argc, argv);
#endif
  if( TOPC_OPT_verbose > 1 ) {
    char buf[61];
    int i;

    printf("\n");
    system("date");
    system("uptime");
    start_time = time(NULL);
    if ( 0 == gethostname(buf, 60) )
      printf("hostname:  %.60s\n", buf);
    else
      printf("hostname:  <gethostname() failed>\n");
    getcwd(buf, 60);
    printf("current directory:  %.60s\n", buf);
    printf("safety level:  %d\n", TOPC_OPT_safety);
    if (TOPC_OPT_safety != 0)
      printf("  safety: >=0: all; >=4: no TOPC_MSG_PTR; >=8: no aggreg.;\n"
             "  >=12: no TOPC_abort_tasks; >=14: no receive thread on slave;\n"
             "  ...;\n"
             "   =19: only 2 slaves; >=20: only 1 slave\n"
             "  (NOT YET FULLY IMPLEMENTED)\n");
    printf("memory model:  %s\n",
	 (COMM_is_shared_memory() ? "shared memory" : "distributed memory"));
    if (!COMM_is_shared_memory())
      printf("procgroup file:  %s\n", TOPC_OPT_procgroup);
    printf("number of slaves:  %d\n", COMM_node_count() - 1);
    /* This is copied from comm.c; abstract out as separate routine? */
    /* Use autoconf AC_REPLACE to replace missing
       pthread_getconcurrency() on Solaris, etc. */
#ifdef HAVE_PTHREAD_SETCONCURRENCY
    if (COMM_is_shared_memory()) {
      /* threads:  slaves + 1  (no thread should need to context switch) */
      printf("number of threads (master+slaves+stubs):  %d\n",
  	     pthread_getconcurrency() );
    }
#endif
    printf("CALLED AS: ");
    for (i = 0; saved_args[i] != NULL && i < ARGS_MAX; i++)
      printf(" %s", saved_args[i]);
    printf("\n\n");
  }
}

/* SEE WHAT autoconf SAYS ABOUT getrusage(), ranlib() */
void TOPC_OPT_finalize( void ) {
  if( TOPC_OPT_verbose > 1 ) {
    struct rusage buf;

    /* getrusage is BSD */
    if ( getrusage( RUSAGE_SELF, &buf ) )
      perror("getrusage");

    printf("\nMASTER:\n");
    printf("Maximum resident set size: %ld\n", buf.ru_maxrss);
    printf("Page faults: %ld\n", buf.ru_majflt);
    printf("Swaps: %ld\n", buf.ru_nswap);
    printf("Voluntary context switches: %ld\n", buf.ru_nvcsw);
    printf("Involuntary context switches: %ld\n", buf.ru_nivcsw);
    printf("CPU time (user s): %ld\n",
           (long)buf.ru_utime.tv_sec /* + buf.ru_utime.tv_usec/1000000 */);

    /* getrusage is BSD */
    if ( getrusage( RUSAGE_CHILDREN, &buf ) )
      perror("getrusage");

    printf("\nSLAVE:\n");
    printf("Maximum resident set site: %ld\n", buf.ru_maxrss);
    printf("Page faults: %ld\n", buf.ru_majflt);
    printf("Swaps: %ld\n", buf.ru_nswap);
    printf("Voluntary context switches: %ld\n", buf.ru_nvcsw);
    printf("Involuntary context switches: %ld\n", buf.ru_nivcsw);
    printf("CPU time (user s): %ld\n",
           (long)buf.ru_utime.tv_sec /* + buf.ru_utime.tv_usec/1000000 */);

    printf("\nParallel computation finished\n");
    system("date");
    system("uptime");
    printf("Elapsed time (s):  %d\n", (int)(time(NULL) - start_time) );
    printf("\n");

    /* We should also call TOPC_master_slave() with do_task() to collect
     * slave statistics.  Do we need to declare do_task() volatile?
     */
  }
}

#ifdef TEST
int main( int argc, char **argv ) {
  TOPC_OPT_pre_init( &argc, &argv );
  TOPC_OPT_pre_init( &argc, &argv );
  exit(0);
}
#endif
