 /**********************************************************************
  * TOP-C (Task Oriented Parallel C)                                   *
  * Copyright (c) 2000 Gene Cooperman <gene@ccs.neu.edu>               *
  *                    and Victor Grinberg <victor@ccs.neu.edu>        *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact Gene Cooperman          *
  * <gene@ccs.neu.edu>.                                                *
  **********************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <stdlib.h> // malloc(), atoi()
#include <string.h> // memcpy(), strlen()
#include <unistd.h> // getcwd(), chdir(), sleep()
#include <assert.h>
#include <stdarg.h>
#include <errno.h>
#include "topc.h"   // data structures shared with application
#include "comm.h"   // not visible to applic.; shared with communication layer

//Compile-time parameters
// Also:  comm.h defines TOPC_MAX_SLAVES and MSG_CACHE_SIZE
#define TOPC_MAX_SLAVES            1000       /* Max. number of slaves */
#define UPDATE_QUEUE_SIZE          200        //Size of update ueue
#define NUM_REUSABLE_IDS           200000000  //Should be >> number of slaves
#define MSG_CACHE_SIZE             10000      //Larger updates are queued
#define MAX_PATH_LEN               256        //for getcwd()

//For ease of further porting
#ifndef FALSE
enum {FALSE, TRUE};
#endif

//Global variables
static int num_slaves;                     //Total number of nodes less 1
static int num_free_slaves;                //Slaves currently not doing a task
static int num_tasks, num_redos, num_updates;  //Statistics
static int last_slave;                     //Slave to which a task was sent last
static int last_task_id;                   //Last task sent to a slave
static int last_update_id;                 //Task that caused last update
static int slave_being_checked;            //Slave whose reply is being checked

//Slave array
struct SLAVE_RECORD {
  TOPC_BOOL   busy;         //Is it doing a task?
  int    msg_cache_free;    //Free bytes in slave's msg_cach
  int    last_update_seen;  //ID of last seen queued update
  int    task_id;           //Task being done
  int    update_id;         //Last update task has seen
  TOPC_BUF input;           //Task input being processed
};
static struct SLAVE_RECORD *slave_array = NULL;
static size_t slave_array_size;            //Number of elements in slave array

//Globals for shared memory
static TOPC_BOOL is_shared_memory = 0;     //Set to 1 by slave, used by master

/***********************************************************************
 ***********************************************************************
 ** MPI tags:  (enum TAG) now defined in comm.h
 ***********************************************************************
 ***********************************************************************/


/***********************************************************************
 ***********************************************************************
 ** Printout functions
 ***********************************************************************
 ***********************************************************************/
// Maybe we can get rid of this now that TOPC_OPT_trace exists.
static void TRACE(char *format,...) {
#if 0
  va_list ap;
  va_start(ap, format);
  vprintf(format, ap);
  printf("\n");
  fflush(stdout);
  va_end(ap);
#endif
}

static void WARNING(char *format,...) {
  va_list ap;
  if (TOPC_OPT_verbose > 0) {
    va_start(ap, format);
    fprintf(stderr, "TOP-C WARNING: ");
    vfprintf(stderr, format, ap);
    fprintf(stderr, "\n");
    va_end(ap);
  }
}

//This function should perhaps take steps to perform proper shutdown
static void ERROR(char *format,...) {
  va_list ap;
  va_start(ap, format);
  fprintf(stderr, "TOP-C: ");
  vfprintf(stderr, format, ap);
  fprintf(stderr, "\n");
  va_end(ap);
  exit(1);
}


/***********************************************************************
 ***********************************************************************
 ** TOP-C task and action interfaces (see topc.h)
 ***********************************************************************
 ***********************************************************************/

//User functions return TOPC_ACTION structs by value, so these ID
//codes are used to identify special return values
enum ACTION {
  NO_ACTION_ID, REDO_ID, UPDATE_ID, CONTINUE_ID
};

TOPC_ACTION NO_ACTION = {NO_ACTION_ID, "NO_ACTION", NULL, 0};
TOPC_ACTION REDO      = {REDO_ID, "REDO", NULL, 0};
TOPC_ACTION UPDATE    = {UPDATE_ID, "UPDATE", NULL, 0};
//Function to form continuation TOPC_ACTION
TOPC_ACTION CONTINUATION(TOPC_BUF msg) {
  TOPC_ACTION action;
  action.type              = CONTINUE_ID;
  action.name              = "CONTINUATION";
  action.continuation      = msg.data;
  action.continuation_size = msg.data_size;
  return action;
}

//Special return values
// NOTE:  dummy != NULL, so user TOPC_MSG(NULL,0) not confused with NOTASK
static char dummy[] = "NOTASK";
TOPC_BUF NOTASK   = {dummy, 0};

#ifndef NO_STACK_CHECK
//Stack marker
static void *stack_bottom = NULL;
#endif

//Function to form a TOPC_BUF return value
TOPC_BUF TOPC_MSG(void *data, size_t data_size) {
  TOPC_BUF msg;
  // slave may return TOPC_MSG(NULL,0) for output
  if ( (data == NULL || data_size == 0) && TOPC_is_master() )
    WARNING("TOPC_MSG() called with null pointer or zero size");

#ifndef NO_STACK_CHECK
 // In pthread, each thread has different stack;
 // A common stack_bottom doesn't work.
 // One could define COMM_stack_bottom() or COMM_is_on_stack().
 // Is there a cleaner patch?
 if ( ! is_shared_memory )
  if (stack_bottom != NULL &&
      (data >= (void*)&msg && data <= stack_bottom ||
       data <= (void*)&msg && data >= stack_bottom))
    WARNING("TOPC_MSG() was passed a pointer into stack on %s.\n%s",
            ( COMM_rank() == 0 ? "master" : "slave" ),
            "  Data buffer should be either static or global.");
#endif

  msg.data      = data;
  msg.data_size = data_size;
  return msg;
}


/***********************************************************************
 ***********************************************************************
 ** Task and updates are identified by reusable IDs, like PIDs on UNIX
 ***********************************************************************
 ***********************************************************************/

/***********************************************************************
 * Compare IDs with respect to rollover.  For ease of initialization,
 * a negative value of first argument compares less then anything
 */
static int idcmp(int id1, int id2) {
  if (id1 == id2) return 0;
  if (id1 < id2 || id1 - id2 > NUM_REUSABLE_IDS / 2) return -1;
  return 1;
}

/***********************************************************************
 * Generate new task ID
 */
static int next_task_id() {
  if (++last_task_id == NUM_REUSABLE_IDS) last_task_id = 0;
  return last_task_id;
}


/***********************************************************************
 ***********************************************************************
 ** Synchronization (atomic I/O) and recording shared data page accesse
 ***********************************************************************
 ***********************************************************************/

static volatile TOPC_BOOL user_handles_own_synchronization = FALSE;

// PAGE_is_up_to_date()

void TOPC_BEGIN_ATOMIC_READ(int pagenum) {
    user_handles_own_synchronization = TRUE;
    // PAGE_atomic_read(pagenum, TOPC_rank());
    COMM_begin_atomic_read();
}
void TOPC_END_ATOMIC_READ(int pagenum) {
    COMM_end_atomic_read();
}
// TOPC_ATOMIC_READ(pagenum)   defined in topc.h

void TOPC_BEGIN_ATOMIC_WRITE(int pagenum) {
    user_handles_own_synchronization = TRUE;
    // PAGE_atomic_write(pagenum, TOPC_rank(), last_update_id);
    COMM_begin_atomic_write();
}
void TOPC_END_ATOMIC_WRITE(int pagenum) {
    COMM_end_atomic_write();
}
// TOPC_ATOMIC_WRITE(pagenum)   defined in topc.h

/***********************************************************************
 ***********************************************************************
 ** Define wrappers for callbacks:  protected stack call, agglomeration
 ** (much of this is only in experimental version)
 ***********************************************************************
 ***********************************************************************/

/***********************************************************************
 * Define wrappers for callback functions
 *   do_task_wrapper/COMM_do_task() as wrapper that takes
 *   agglomerated input and repeatedly calls application-defined do_task()
 *   to produce agglomerated output for output.data
 */
static TOPC_BUF (*generate_task_input_orig)(void);
static TOPC_BUF (*do_task_orig)(void *input);
static TOPC_ACTION (*check_task_result_orig)(void *input, void* output);
static void (*update_shared_data_orig)(void *input, void *output);

static TOPC_BUF generate_task_input_wrapper() {
    return (*generate_task_input_orig)();
}
static TOPC_BUF do_task_wrapper(void *input) {
  return (*do_task_orig)(input);
}
static TOPC_ACTION check_task_result_wrapper(void *input, void *output) {
    return (*check_task_result_orig)(input, output);
}
static void update_shared_data_wrapper(void *input, void *output) {
    (*update_shared_data_orig)(input, output);
}

//Application-defined callbacks
//The communication layer may redefine them
//  (e.g.:  COMM_init() may add another wrapper)
TOPC_BUF (*COMM_generate_task_input)(void) = generate_task_input_wrapper;
TOPC_BUF (*COMM_do_task)(void *input) = do_task_wrapper;
TOPC_ACTION (*COMM_check_task_result)(void *input, void *output)
	= check_task_result_wrapper;
void (*COMM_update_shared_data)(void *input, void *output)
	= update_shared_data_wrapper;
//Can be called by communication layer
// Assigned to slave_loop() in TOPC_master_slave().
void (*COMM_slave_loop)(void) = NULL;


/***********************************************************************
 ***********************************************************************
 ** Update queue on master
 ***********************************************************************
 ***********************************************************************/

//Assume update messages are either all short or all long.
//If no update messages pending, all messages are short,
//  and msg_cache for each slave has space, then
//  immediately broadcast them, using the network as our buffer.
//Otherwise, queue the messages, and then update_slave() sends
//  all pending updates to a slave just before sending the
//  next task input (whether due to REDO, CONTINUATION, or new task)
//PERHAPS WE SHOULD ALSO SEND PENDING UPDATES TO A SLAVE IMMEDIATELY UPON SLAVE
//  RETURN WITHOUT WAITING TO FIRST CALL check_task_result()
//  AND POSSIBLY DO UPDATE ON MASTER.
//If more slaves return while the master was doing its own update,
//  then those slaves will see all updates when the master recognizes them.
//Note that an urgent message (e.g.: SOFT_ABORT), can always be
//  probed by slave, since all updates fit in slave msg_cache.

//Update queue
struct UPDATE_RECORD {
  int    id;              //ID of task that resulted in the update
  void  *input;           //Update information
  size_t input_size;
  void  *output;
  size_t output_size;
  int    slaves_updated;  //Number of slaves this update has been sent to
};
static struct UPDATE_RECORD update_queue[UPDATE_QUEUE_SIZE],
                            *update_queue_tail, *update_queue_head;

void free_input_data(void *ptr) { free(ptr); }
// More efficient version for now:
#define free_input_data(x) (free(x))

/***********************************************************************
 * Send update message to a given slave
 */
static void send_update(int slave, void *input, size_t input_size,
                        void *output, size_t output_size) {
  COMM_send_msg(input, input_size, slave, UPDATE_INPUT_TAG);
  COMM_send_msg(output, output_size, slave, UPDATE_OUTPUT_TAG);
}

/***********************************************************************
 * Broadcast update message
 */
static void broadcast_update(void *input, size_t input_size,
                             void *output, size_t output_size) {
  int slave;
  for (slave = 1; slave <= num_slaves; ++slave) {
    send_update(slave, input, input_size, output, output_size);
    slave_array[slave].msg_cache_free -=
      (input_size + output_size + 2*sizeof(enum TAG) + 2*sizeof(input_size));
  }
}

// COULD USE getsockopt(,,SO_RCVBUF,result,&result_size);
//  TO GET OR CHANGE DEFAULT SOCKET BUFFER SIZE FOR MORE ACCURACY.
// IF SIZE IS MORE THAN NETWORK BUFFER SIZE, THEN CONSIDER INCREASING
//  SO_SNDBUF and SO_RCVBUF.
// THIS COULD ALSO BE USEFUL FOR aggregation/agglomeration FOR DIST. MEMORY.
// PREFER WORD aggregation?

static TOPC_BOOL broadcast_will_not_block(int input_size, int output_size) {
  int slave;
  for (slave = 1; slave <= num_slaves; ++slave)
    if (input_size + output_size + 2*sizeof(enum TAG) + 2*sizeof(input_size)
        > slave_array[slave].msg_cache_free)
      return FALSE;
  return TRUE;
}

/***********************************************************************
 * Find first update that the slave hasn't seen.  If there are no
 * unseen updates, return value points past the end of the queue
 */
static struct UPDATE_RECORD *find_first_unseen_update(int slave) {
  struct UPDATE_RECORD *l, *m, *r, *mm;

  l = update_queue_head;
  r = update_queue_tail;
  if (r < l) r += UPDATE_QUEUE_SIZE;

  while (r > l) {
    m = l + (r - l) / 2;
    mm = m < update_queue + UPDATE_QUEUE_SIZE ? m : m - UPDATE_QUEUE_SIZE;
    if (idcmp(slave_array[slave].last_update_seen, mm->id) < 0)
      r = m;
    else
      l = m + 1;
  }
  assert(l == r);
  return r < update_queue + UPDATE_QUEUE_SIZE ? r : r - UPDATE_QUEUE_SIZE;
}

/***********************************************************************
 * Send all necessary updates to a slave.  Free memory occupied by
 * updates that have been sent to all slaves
 */
static void update_slave(int slave) {
  struct UPDATE_RECORD *update;
  int slaves_updated;

  //Send all unseen updates to the slave.  Increment updated slave
  //count for each update.  When an update has been sent to all
  //slaves, remove it from queue.  The removal should only happen at
  //the end of the queue, because updates are sent in order
  update = find_first_unseen_update(slave);
  while (update != update_queue_tail) {
    send_update(slave, update->input, update->input_size, update->output,
                update->output_size);
    slave_array[slave].last_update_seen = update->id;

    slaves_updated = ++update->slaves_updated;
    assert(slaves_updated < num_slaves || update == update_queue_head);

    if (++update == update_queue + UPDATE_QUEUE_SIZE)
      update = update_queue;
    if (slaves_updated == num_slaves) {
      free_input_data(update_queue_head->input);
      COMM_free_receive_buf(update_queue_head->output);
      update_queue_head = update;
    }
  }
  //update_slave() always called when slave finishes task
  slave_array[slave].msg_cache_free = MSG_CACHE_SIZE;
}

/***********************************************************************
 * Queue an update for being sent to all slaves eventually.  The
 * update queue must have space on it
 */
static void queue_update(int id, void *input, size_t input_size,
                         void *output, size_t output_size) {
  update_queue_tail->id             = id;
  update_queue_tail->input          = input;
  update_queue_tail->input_size     = input_size;
  update_queue_tail->output         = output;
  update_queue_tail->output_size    = output_size;
  update_queue_tail->slaves_updated = 0;
  if (++update_queue_tail == update_queue + UPDATE_QUEUE_SIZE)
    update_queue_tail = update_queue;
  if (update_queue_head == update_queue_tail)
    ERROR("Update queue overflow");
}

/***********************************************************************
 * Schedule global update.  Long updates are queued.  Short updates
 * are only queued if there is something on the queue already, to make
 * sure updates are not reordered.  Unless update is queued, frees
 * memory
 */
static void schedule_update(int id, void *input, size_t input_size,
                            void *output, size_t output_size) {
  if (broadcast_will_not_block(input_size, output_size)
      && update_queue_head == update_queue_tail) {
    broadcast_update(input, input_size, output, output_size);
    free_input_data(input);
    COMM_free_receive_buf(output);
  }
  else
    queue_update(id, input, input_size, output, output_size);
  last_update_id = id;
}

/***********************************************************************
 ***********************************************************************
 ** Tracing
 ***********************************************************************
 ***********************************************************************/

/***********************************************************************
 * Default:  trace msg; value of NULL: print nothing
 * User can redefine initial binding to print buffer-specific tracing msg
 */

static void do_nothing(){};
void (*TOPC_OPT_trace_input)() = do_nothing;  /* args: (void *input) */
void (*TOPC_OPT_trace_result)() = do_nothing;
  /* args: (void *input, void *output) */

static void trace_input(void *input, int slave) {
  if (TOPC_OPT_trace == FALSE) return;
  else if (input == NOTASK.data) { printf("master: NOTASK\n"); return; }
  else { printf("master -> %d: ", slave);
         if (TOPC_OPT_trace == 2 && *TOPC_OPT_trace_input)
           (*TOPC_OPT_trace_input)(input);
         printf("\n");
       }
}
static void trace_result(int slave, void *input, void *output) {
  if (TOPC_OPT_trace == FALSE) return;
  else { printf("%d -> master: ", slave);
         if (TOPC_OPT_trace == 2 && TOPC_OPT_trace_result)
           (*TOPC_OPT_trace_result)(input, output);
         printf("\n");
       }
}
static void trace_action(TOPC_ACTION action, void *data, void *out, int slave) {
  if (TOPC_OPT_trace == FALSE) return;
  if (action.type == REDO_ID) { printf("REDO: "); trace_input(data, slave); }
  else if (action.type == UPDATE_ID)
    { printf("  UPDATE: ");
      if (TOPC_OPT_trace == 2)
        (*TOPC_OPT_trace_result)(data, out);
      printf("\n");
    }
  else if (action.type == CONTINUE_ID)
    { printf("  CONTINUATION: "); trace_input(data, slave); }
}

/***********************************************************************
 ***********************************************************************
 ** Master-slave processing
 ***********************************************************************
 ***********************************************************************/

/***********************************************************************
 * Slave processing loop (slave)
 */
static void slave_loop() {
  enum TAG tag;
  void *input, *output;
  size_t input_size, output_size;
  TOPC_BUF result;

  is_shared_memory = 1;  // Initialized to 0; Master reads 1 iff shared memory
  num_tasks = num_updates = 0;
  for (;;) {
    input = NULL; /* Tell COMM layer to provide buffer */
    COMM_receive_msg(&input, &input_size, NULL, &tag);
    switch (tag) {
    case UPDATE_INPUT_TAG:
      output = NULL; /* Tell COMM layer to provide buffer */
      COMM_receive_msg(&output, &output_size, NULL, &tag);
      if (tag != UPDATE_OUTPUT_TAG)
        ERROR("Received %d instead of UPDATE_OUTPUT_TAG");

      ++num_updates;
      (*COMM_update_shared_data)(input, output);
      COMM_free_receive_buf(input);
      // malloc(0) can return NULL instead of a 0-length buffer
      if (output) COMM_free_receive_buf(output);
      break;

    case TASK_INPUT_TAG:
      ++num_tasks;
      COMM_abort_pending = 0;  //Turn off aborts before new task, REDO, or CONT
      result = (*COMM_do_task)(input);
      COMM_free_receive_buf(input);
      COMM_send_msg(result.data, result.data_size, 0, SLAVE_REPLY_TAG);
      break;

    case END_MASTER_SLAVE_TAG:
      return;

    case PING_TAG:
      COMM_send_msg(NULL, 0, 0, PING_TAG);
      break;

    case CHDIR_TAG:
      if (0 != chdir((char *)input))
        WARNING("SLAVE %d:  Couldn't change directory to:\n  %s",
                TOPC_rank(), (char *)input);
      COMM_free_receive_buf(input);
      break;

    case SLAVE_WAIT_TAG:
      sleep(*(int *)input);
      COMM_free_receive_buf(input);
      break;

    default:
      ERROR("Received invalid tag %d", tag);
    }
  }
}

/***********************************************************************
 * Find available slave.  Should only be called when it's known that
 * there is one (master)
 */
static int find_available_slave() {
  int i;

  //Loop starts where we left off last time, to spread the load evenly
  //across all slaves
  for (i = 0; i < num_slaves; ++i) {
    if (++last_slave > num_slaves) last_slave = 1;
    if (!slave_array[last_slave].busy) break;
  }
  assert(!slave_array[last_slave].busy);
  return last_slave;
}

/***********************************************************************
 * Write barrier:  Wait for all slaves to finish tasks (master)
 */
static void wait_for_all_slaves() {
  int rank = 1;
  for (rank = 1; rank <= num_slaves; rank++) {
    if (slave_array[rank].busy)
      COMM_probe_msg_blocking(rank);
  }
}

/***********************************************************************
 * Send task input, REDO, or CONTINUATION to slave (master)
 */
static void send_task_input(int slave, TOPC_BUF input) {
  //Send all pending updates to the slave
  update_slave(slave);

  //Record task information and send the task
  slave_array[slave].busy       = TRUE;
  //gdc - incrementing task_id here means REDO and CONTINUATION also get
  //      new task_id's.  Is this the desired behavior?
  slave_array[slave].task_id    = next_task_id();
  slave_array[slave].update_id  = last_update_id;
  COMM_send_msg(input.data, input.data_size, slave, TASK_INPUT_TAG);
}

/***********************************************************************
 * Receive and process output of a slave computation (master)
 */
static void receive_task_output() {
  TOPC_ACTION action;
  TOPC_BUF msg_buf;
  void *output = NULL;
  size_t output_size;
  int slave;

  //Receive task output from slave, check it
  COMM_receive_msg(&output, &output_size, &slave, NULL);
  trace_result(slave, slave_array[slave].input.data, output);
  slave_being_checked = slave;
  action = (*COMM_check_task_result)(slave_array[slave].input.data, output);
  slave_being_checked = 0;
  TRACE("Reply from %d", slave);

  //Take required action
  switch (action.type) {
  case NO_ACTION_ID:
    slave_array[slave].busy = FALSE;
    ++num_free_slaves;
    TRACE("NO_ACTION");
    COMM_free_receive_buf(output);
    free_input_data(slave_array[slave].input.data);
    break;

  case REDO_ID:
    trace_action(action, slave_array[slave].input.data, output, slave);
    COMM_free_receive_buf(output);
    msg_buf.data = slave_array[slave].input.data;
    msg_buf.data_size = slave_array[slave].input.data_size;
    send_task_input(slave, msg_buf);
    ++num_redos;
    break;

  case UPDATE_ID:
    trace_action(action, slave_array[slave].input.data, output, slave);
    slave_array[slave].busy = FALSE;
    ++num_free_slaves;
    TRACE("UPDATE");
    // If shared memory, default is that UPDATE acts as write barrier
    if ( is_shared_memory && ! user_handles_own_synchronization )
      wait_for_all_slaves();
    // Call application callback on master
    (*COMM_update_shared_data)(slave_array[slave].input.data, output);
    if ( is_shared_memory ) free_input_data(slave_array[slave].input.data);
    else schedule_update( slave_array[slave].task_id,
                          slave_array[slave].input.data,
                          slave_array[slave].input.data_size,
		          output, output_size);
    ++num_updates;
    break;

  case CONTINUE_ID:
    trace_action(action, action.continuation, output, slave);
    COMM_free_receive_buf(output);
    msg_buf.data = action.continuation;
    msg_buf.data_size = action.continuation_size;
    send_task_input(slave, msg_buf);
    break;

  default:
    ERROR("check_task_result() returned invalid action %d", action.type);
  }
}

/***********************************************************************
 * Receive and process result = (input,output) of a slave computation (master)
 */
static void wait_for_free_slave() {
  //NOTE:  COMM_probe_msg() is non-blocking.
  //NOTE:  receive_task_output() blocking due to call to COMM_receive_msg().
  while (COMM_probe_msg() || num_free_slaves == 0)
    receive_task_output();
  assert( num_free_slaves > 0 );
}

/***********************************************************************
 * Submit task for dispatching to a slave.  Returns FALSE only when no
 * slaves are busy, so there's no chance of getting more tasks from
 * application code (master)
 */
static TOPC_BOOL submit_task_input(TOPC_BUF input) {
  int slave;

  slave = find_available_slave();

  //Copy from application space to TOP-C space
  slave_array[slave].input.data = malloc(input.data_size);
  if (slave_array[slave].input.data == NULL)
    ERROR("Out of memory trying to copy task input");
  memcpy(slave_array[slave].input.data, input.data, input.data_size);
  slave_array[slave].input.data_size = input.data_size;

  // trace_input must come after receive_task_output()
  trace_input(input.data, slave);
  send_task_input(slave, slave_array[slave].input);
  --num_free_slaves;
  ++num_tasks;
  return TRUE;
}

/***********************************************************************
 * Initialize master (master)
 */
static void master_init() {
  int slave;

  //Initialize statistics and run parameters
  num_tasks = num_redos = num_updates = 0;
  num_free_slaves = num_slaves = COMM_node_count() - 1;
  last_slave     = 0;   //No last slave
  last_task_id   = -1;  //No last task ID
  last_update_id = -1;  //No updates seen so far
  slave_being_checked = 0;  //Not inside check_task_result()

  //Allocate and initialize slave array
  if (slave_array && (signed int)slave_array_size != num_slaves + 1) {
    free(slave_array);
    slave_array = NULL;
  }
  if (!slave_array) {
    slave_array_size = num_slaves + 1;
    slave_array = calloc(slave_array_size, sizeof(struct SLAVE_RECORD));
    if (slave_array == NULL) ERROR("No memory for process table");
  }
  for (slave = 1; slave <= num_slaves; ++slave) {
    slave_array[slave].last_update_seen = -1;  //Seen none yet
    slave_array[slave].busy = FALSE;
    slave_array[slave].msg_cache_free = MSG_CACHE_SIZE;
  }

  //Initialize update queue
  update_queue_head = update_queue_tail = update_queue;
}

/***********************************************************************
 * Force slaves to change to same directory as master (ignore if no such dir)
 */
static void master_slave_chdir() {
  if ( ! COMM_initialized() )
    ERROR("master_slave_chdir() called before TOPC_init()");
  // If not shared memory, ask slave to change to same directory as master
  if ( ! COMM_is_shared_memory() ) {
    if (TOPC_is_master()) {
      int slave;
      char cwd[MAX_PATH_LEN], *cwd_ptr;

      cwd_ptr = getenv("PWD"); /* prefer logical path by which we got here */
      if (!cwd_ptr) cwd_ptr = getcwd(cwd, MAX_PATH_LEN);
      // Emulate master side of master_slave loop
      for (slave = 1; slave <= num_slaves; ++slave) {
        COMM_send_msg(cwd_ptr, 1+strlen(cwd_ptr), slave, CHDIR_TAG);
        if (TOPC_OPT_slave_wait > 0)
          COMM_send_msg(&TOPC_OPT_slave_wait, sizeof(TOPC_OPT_slave_wait),
                        slave, SLAVE_WAIT_TAG);
        COMM_send_msg(NULL, 0, slave, END_MASTER_SLAVE_TAG);
      }
    }
    // each slave sees CHDIR_TAG, END_MASTER_SLAVE_TAG (then exits slave loop)
    else slave_loop();
  }
}


/***********************************************************************
 ***********************************************************************
 ** Interface (public) functions
 ***********************************************************************
 ***********************************************************************/

/***********************************************************************
 * Get node count (pass to comm layer)
 */
int TOPC_node_count() {
  return COMM_node_count();
}

/***********************************************************************
 * Get number of slaves
 */
int TOPC_num_slaves() {
  return COMM_node_count() - 1;
}

/***********************************************************************
 * Get rank of this node (pass to comm layer)
 */
int TOPC_rank() {
  return COMM_rank();
}

/***********************************************************************
 * Figure out who is the master
 */
TOPC_BOOL TOPC_is_master() {
  return COMM_rank() == 0;
}

/***********************************************************************
 * Initialize TOP-C and COMM (MPI, TOPC_OPTION, etc.) (both master and slave)
 */
// old TOPC_OPTION(var,val) syntax for TOPC_init() is obsolete:
// (struct opt_list)  defined in comm.h
static struct opt_list options_list = { 0 }; // init: options_list.count = 0
#define OPTIONS_COUNT ( options_list.count )

struct TOPC_OPT TOPC_OPTION(char *n, ...) {
  va_list ap;
  int i;

  va_start(ap, n);
  printf("*** WARNING:  TOPC_OPTION(var,val) is obsolete.\n"
         "***  Please use the new form:   TOPC_OPT_var=val;\n"
         "***  Do ./a.out --TOPC_help    for a current list of parameters,\n"
         "***    where --TOPC_var=val, also has a form:  TOPC_OPT_var=val;\n");
  i = OPTIONS_COUNT++;
  options_list.options[i].name = n;
  if ( 0 == strncmp( options_list.options[i].name, "trace", 20 )
       || 0 == strncmp(options_list.options[i].name, "num_slave_threads", 20) )
    options_list.options[i].value.integer = va_arg(ap,int);
  if ( 0 == strncmp( options_list.options[i].name, "trace_input", 20 )
       || 0 == strncmp( options_list.options[i].name, "trace_result", 20 )
       || 0 == strncmp( options_list.options[i].name, "trace_action", 20 ) ) {
    options_list.options[i].value.fnc = va_arg(ap,TOPC_FUNCTION *);
  }
  va_end(ap);
}

/*
doc:
    A TOP-C variable has either global scope or local scope.
    It is also shared or private.
    A variable is global or local according to the standard C specification.
    In the distributed memory (e.g.: -mpi), a variable is always private.
      Even a variable whose value is in shared data is private, but
      calls to UpdateSharedData will update the private value with
      the latest value.  Thus, we say that such variables have lazy updates.
    In the sequential memory (e.g.: -seq), a variable is always shared.
      Thus UpdateSharedMemory executes only once, in the sequential process.
    In shared memory (e.g.: -pthread), a variable is shared by default.
      Further UpdateSharedMemory executes only once, in the master thread.
      However, certain strategies for high concurrency require the use
      of private variables to privately save temporary values and use them
      again during a REDO action.
      To accomodate this, one can declare a single private variable.
      The private variable has global scope, but each thread has a separate
      copy that can be read or written only by that thread.
      One is allowd only a single private variable, but its type can
      be declared arbitrarily.  Hence, in the example below, we declare
      TOPC_private_global to be a struct, and then use
      TOPC_private_global.val1 and TOPC_private_global.val2 separately.

// Usage:
// TOPC_declare_private_global( struct {int val1; float val2;} );
// void foo() {
//   TOPC_private_global.val1 = 42;
//   TOPC_private_global.val2 = 3.14;
// }
// void bar() {
//   foo();
//   if (TOPC_private_global.val1 != 42) printf("ERROR");
//   if (TOPC_private_global.val2 != 3.14) printf("ERROR");
// }
*/

void TOPC_init(int *argc, char ***argv, ...) {
  int i;
  void *tmp;
  va_list ap;

  errno = 0;
  tmp = malloc(1000);
  if (tmp == NULL && errno != 0) {
    perror("TOP-C initialization:");
    if (errno == ENOMEM)
      printf("This can happen if your application dynamically OR statically\n"
             "  allocates a lot of memory.  For example,\n"
             "    char *x[100000000];\n"
             "  might cause this error, depending on your swap size.\n");
    exit(1);
  }
  else
    free(tmp);
  va_start(ap, argv);
  for ( i = 0; i < OPTIONS_COUNT; i++ ) {
    if ( 0 == strncmp( options_list.options[i].name, "trace", 20 ) )
      TOPC_OPT_trace = options_list.options[i].value.integer;
    if ( 0 == strncmp( options_list.options[i].name, "trace_input", 20 ) )
      TOPC_OPT_trace_input = options_list.options[i].value.fnc;
    if ( 0 == strncmp( options_list.options[i].name, "trace_result", 20 ) )
      TOPC_OPT_trace_result = options_list.options[i].value.fnc;
  }

  // In distributed memory, this is also called on slave; should be harmless.
  TOPC_OPT_pre_init( argc, argv );

  COMM_init(argc, argv, &options_list);
  OPTIONS_COUNT = 0; // reset to zero options
  va_end(ap);

  is_shared_memory = COMM_is_shared_memory();
  if (TOPC_is_master()) master_init();
  master_slave_chdir();  // Place slaves in same directory as master

  if ( TOPC_is_master() )
    TOPC_OPT_post_init( argc, argv );
}

/***********************************************************************
 * Shutdown TOP-C and MPI (both master and slave)
 */
void TOPC_finalize() {
  if ( TOPC_is_master() )
    TOPC_OPT_finalize();
  COMM_finalize();
}

/***********************************************************************
 * Was the shared memory up to date?  Should only be called from within
 * check_task_result()
 */
TOPC_BOOL TOPC_is_up_to_date() {
  int cmp;

  if (slave_being_checked == 0)
    ERROR("up_to_date() called from outside of check_task_result()");

  cmp = idcmp(slave_array[slave_being_checked].update_id, last_update_id);
  assert(cmp <= 0);
  return cmp == 0;
}

/***********************************************************************
 * Parallel execution.  This function is called on both master and
 * slave, and executes the parallel code supplied in the form of four
 * callback routines (both master and slave)
 */
void TOPC_master_slave(
  TOPC_BUF (*generate_task_input_)(void),
  TOPC_BUF (*do_task_)(),              /* args: (void *input) */
  TOPC_ACTION (*check_task_result_)(), /* args: (void *input, void *output) */
  void (*update_shared_data_)()        /* args: (void *input, void *output) */
) {
  TOPC_BUF input;
  int slave;

#ifndef NO_STACK_CHECK
  //Mark stack for error checking in TOPC_MSG().  See comment in
  //TOPC_raw_begin_master_slave() for usage
  stack_bottom = &input;
#endif

  if (!COMM_initialized()) ERROR("TOPC_master_slave called before TOPC_init()");

  //Store callbacks into global variables.  If we want to support
  //  recursion these need to go on a stack
  // xxx_orig() is called by COMM_xxx for xxx a callback function.
  generate_task_input_orig  = generate_task_input_;
  do_task_orig              = do_task_;
  check_task_result_orig    = check_task_result_;
  update_shared_data_orig   = update_shared_data_;
  COMM_slave_loop           = slave_loop; // needed for pthreads/SMP COMM layer

  //Master loop --- heart of the TOP-C algorithm
  if ( TOPC_is_master() ) {
    do {
      wait_for_free_slave();           // And process any pending slave replies
      input = COMM_generate_task_input();
      if (input.data != NOTASK.data)
        submit_task_input(input);            // lower num_free_slaves
      else if (num_free_slaves < num_slaves) // else insure progress condition
        receive_task_output();               // by blocking until slave reply
    } while (input.data != NOTASK.data || num_free_slaves < num_slaves);
    assert( input.data == NOTASK.data && num_free_slaves == num_slaves );

    //Check if user has not supplied any tasks for this run
    if (num_tasks == 0)
      WARNING("GenerateTaskInput() returned null input on first invocation");

    for (slave = 1; slave <= num_slaves; ++slave)
      COMM_send_msg(NULL, 0, slave, END_MASTER_SLAVE_TAG);
  }

  //Slave loop
  else slave_loop();
}

/***********************************************************************
 ***********************************************************************
 ** Soft abort
 ***********************************************************************
 ***********************************************************************/
void TOPC_abort_tasks() {
  if (!TOPC_is_master())
    ERROR("TOPC_abort_tasks called on a slave");
  if (TOPC_OPT_safety < 12)
    COMM_soft_abort();
}

/* What about in UpdateSharedData() on master? */
/* Should we restrict it to is_in_do_task, and add static version in topc.c? */
TOPC_BOOL TOPC_abort_pending() {
  if (TOPC_is_master())
    WARNING("TOPC_abort_pending:  called on master; result not useful");
  return COMM_abort_pending;
}

/***********************************************************************
 ***********************************************************************
 ** Alternate TOPC_raw master-slave interface:
 **
 **   The idea is that a traditional generate_task_input() is a kind of
 **   iterator that needs to be provided by application code.  However,
 **   sometimes the original sequential code produces task inputs
 **   inside of complicated nested loops.  In such cases, it is
 **   difficult to create a corresponding iterator (and co-routines or
 **   threads would in fact be the ideal language construct).
 **
 **   To get around this, we replace a single call to master_slave() by:
 **     TOPC_raw_begin_master(do_task, check_task_result, update_shared_data)
 **     TOPC_raw_submit_task_input(input)
 **     TOPC_raw_wait_for_task_result()
 **     TOPC_raw_end_master()
 **
 **   The application can then call submit_task_input() repeatedly with
 **   the new task inputs, before completing the computation by a call
 **   to TOPC_end_master_slave().
 **
 **   Slave only calls TOPC_raw_begin_master_slave() and
 **   TOPC_raw_end_master_slave(), whereas slave loop is in the former,
 **   and the latter does nothing
 ***********************************************************************
 ***********************************************************************/

/***********************************************************************
 * Start parallel execution on master
 */
void TOPC_raw_begin_master_slave(
  TOPC_BUF (*do_task_)(),              /* args: (void *input) */
  TOPC_ACTION (*check_task_result_)(), /* args: (void *input, void *output) */
  void (*update_shared_data_)()        /* args: (void *input, void *output) */
) {
#ifndef NO_STACK_CHECK
  int stack_marker;
#endif

  if (!COMM_initialized())
    ERROR("TOPC_raw_begin_master_slave called before TOPC_init()");

  // xxx_orig() is called by COMM_xxx for xxx a callback function.
  do_task_orig              = do_task_;
  check_task_result_orig    = check_task_result_;
  update_shared_data_orig   = update_shared_data_;
  COMM_slave_loop           = slave_loop; // needed for pthreads/SMP COMM layer

  //If the application code calls TOPC_begin_master_slave() from
  //inside some functions, and TOPC_raw_submit_task_input() is
  //outside, we will mark less then the whole stack.  But we will
  //still mark off a section of the stack, so no spurious errors
  //should result.  Though, beware of segmented memory models, where
  //pointer comparisons in different segments may not work as expected
#ifndef NO_STACK_CHECK
  stack_bottom = &stack_marker;
#endif

  if (!TOPC_is_master())
    slave_loop();
}

/***********************************************************************
 * End parallel execution on master
 */
void TOPC_raw_end_master_slave() {
  int slave;

  if (TOPC_is_master()) {
    if (num_tasks == 0)
      WARNING("GenerateTaskInput() returned NOTASK on first invocation");
    while (num_free_slaves < num_slaves)
      receive_task_output();

    for (slave = 1; slave <= num_slaves; ++slave)
      COMM_send_msg(NULL, 0, slave, END_MASTER_SLAVE_TAG);
  }
}

/***********************************************************************
 * Submit task input to TOPC for parallel execution
 */
void TOPC_raw_submit_task_input(TOPC_BUF input) {
  if (TOPC_is_master() && input.data != NOTASK.data) {
    wait_for_free_slave();
    submit_task_input(input);
  }
}

/***********************************************************************
 * If pending task exists, wait for it, call CheckTaskResult() and return true
 * If no pending task exists, return false
 */
TOPC_BOOL TOPC_raw_wait_for_task_result() {
  if (num_free_slaves < num_slaves) {
    do {
      receive_task_output();
    } while (COMM_probe_msg() || num_free_slaves == 0);
    return TRUE;
  }
  else return FALSE;
}
